@@ -26,6 +26,7 @@ import qualified Functions
 import LaTeX (LatexContext (LatexContext), explainRules, programToLaTeX, rewrittensToLatex)
 import Lining (LineFormat (MULTILINE, SINGLELINE))
 import Logger
+import Merge (merge)
 import Misc (ensuredFile)
 import qualified Misc
 import Must (Must (..))
@@ -42,14 +43,21 @@ import XMIR (XmirContext (XmirContext), defaultXmirContext, parseXMIRThrows, pri
 import Yaml (normalizationRules)
 import qualified Yaml as Y
 
+data PrintProgramContext = PrintProgCtx
+  { sugar :: SugarType,
+    line :: LineFormat,
+    xmirCtx :: XmirContext,
+    nonumber :: Bool
+  }
+
 data CmdException
-  = InvalidRewriteArguments {message :: String}
+  = InvalidCLIArguments {message :: String}
   | CouldNotReadFromStdin {message :: String}
   | CouldNotDataize
   deriving (Exception)
 
 instance Show CmdException where
-  show InvalidRewriteArguments {..} = printf "Invalid set of arguments for 'rewrite' command: %s" message
+  show InvalidCLIArguments {..} = printf "Invalid set of arguments: %s" message
   show CouldNotReadFromStdin {..} = printf "Could not read input from stdin\nReason: %s" message
   show CouldNotDataize = "Could not dataize given program"
 
@@ -57,6 +65,7 @@ data Command
   = CmdRewrite OptsRewrite
   | CmdDataize OptsDataize
   | CmdExplain OptsExplain
+  | CmdMerge OptsMerge
 
 data IOFormat = XMIR | PHI | LATEX
   deriving (Eq)
@@ -98,6 +107,7 @@ data OptsRewrite = OptsRewrite
     omitListing :: Bool,
     omitComments :: Bool,
     depthSensitive :: Bool,
+    nonumber :: Bool,
     must :: Must,
     maxDepth :: Integer,
     maxCycles :: Integer,
@@ -108,27 +118,17 @@ data OptsRewrite = OptsRewrite
     inputFile :: Maybe FilePath
   }
 
-parseIOFormat :: String -> ReadM IOFormat
-parseIOFormat type' = eitherReader $ \format -> case (map toLower format, type') of
-  ("xmir", _) -> Right XMIR
-  ("phi", _) -> Right PHI
-  ("latex", "output") -> Right LATEX
-  _ -> Left (printf "The value '%s' can't be used for '--%s' option, use --help to check possible values" format type')
-
-argInputFile :: Parser (Maybe FilePath)
-argInputFile = optional (argument str (metavar "FILE" <> help "Path to input file"))
-
-optMaxDepth :: Parser Integer
-optMaxDepth = option auto (long "max-depth" <> metavar "DEPTH" <> help "Maximum number of rewriting iterations per rule" <> value 25 <> showDefault)
-
-optMaxCycles :: Parser Integer
-optMaxCycles = option auto (long "max-cycles" <> metavar "CYCLES" <> help "Maximum number of rewriting cycles across all rules" <> value 25 <> showDefault)
-
-optInputFormat :: Parser IOFormat
-optInputFormat = option (parseIOFormat "input") (long "input" <> metavar "FORMAT" <> help "Program input format (phi, xmir, latex)" <> value PHI <> showDefault)
-
-optDepthSensitive :: Parser Bool
-optDepthSensitive = switch (long "depth-sensitive" <> help "Fail if rewriting is not finished after reaching max attempts (see --max-cycles or --max-depth)")
+data OptsMerge = OptsMerge
+  { logLevel :: LogLevel,
+    inputFormat :: IOFormat,
+    outputFormat :: IOFormat,
+    sugarType :: SugarType,
+    flat :: LineFormat,
+    omitListing :: Bool,
+    omitComments :: Bool,
+    targetFile :: Maybe FilePath,
+    inputs :: [FilePath]
+  }
 
 optLogLevel :: Parser LogLevel
 optLogLevel =
@@ -152,8 +152,33 @@ optLogLevel =
       "NONE" -> Right NONE
       _ -> Left $ "unknown log-level: " <> lvl
 
+parseIOFormat :: String -> ReadM IOFormat
+parseIOFormat type' = eitherReader $ \format -> case (map toLower format, type') of
+  ("xmir", _) -> Right XMIR
+  ("phi", _) -> Right PHI
+  ("latex", "output") -> Right LATEX
+  _ -> Left (printf "The value '%s' can't be used for '--%s' option, use --help to check possible values" format type')
+
+argInputFile :: Parser (Maybe FilePath)
+argInputFile = optional (argument str (metavar "FILE" <> help "Path to input file"))
+
+optMaxDepth :: Parser Integer
+optMaxDepth = option auto (long "max-depth" <> metavar "DEPTH" <> help "Maximum number of rewriting iterations per rule" <> value 25 <> showDefault)
+
+optMaxCycles :: Parser Integer
+optMaxCycles = option auto (long "max-cycles" <> metavar "CYCLES" <> help "Maximum number of rewriting cycles across all rules" <> value 25 <> showDefault)
+
+optInputFormat :: Parser IOFormat
+optInputFormat = option (parseIOFormat "input") (long "input" <> metavar "FORMAT" <> help "Program input format (phi, xmir)" <> value PHI <> showDefault)
+
+optOutputFormat :: Parser IOFormat
+optOutputFormat = option (parseIOFormat "output") (long "output" <> metavar "FORMAT" <> help "Program output format (phi, xmir, latex)" <> value PHI <> showDefault)
+
+optDepthSensitive :: Parser Bool
+optDepthSensitive = switch (long "depth-sensitive" <> help "Fail if rewriting is not finished after reaching max attempts (see --max-cycles or --max-depth)")
+
 optRule :: Parser [FilePath]
-optRule = many (strOption (long "rule" <> metavar "FILE" <> help "Path to custom rule"))
+optRule = many (strOption (long "rule" <> metavar "[FILE]" <> help "Path to custom rule"))
 
 optNormalize :: Parser Bool
 optNormalize = switch (long "normalize" <> help "Use built-in normalization rules")
@@ -167,11 +192,23 @@ optStepsDir = optional (strOption (long "steps-dir" <> metavar "FILE" <> help "D
 optShuffle :: Parser Bool
 optShuffle = switch (long "shuffle" <> help "Shuffle rules before applying")
 
-optSugar :: Parser SugarType
-optSugar = flag SALTY SWEET (long "sweet" <> help "Print result and intermediate (see --steps flag) 𝜑-programs using syntax sugar")
+optSugar :: [String] -> Parser SugarType
+optSugar opts = flag SALTY SWEET (long "sweet" <> help (printf "Print result and intermediate (see %s option(s)) 𝜑-programs using syntax sugar" (intercalate ", " opts)))
 
-optLineFormat :: Parser LineFormat
-optLineFormat = flag MULTILINE SINGLELINE (long "flat" <> help "Print result and intermediate (see --steps flag) 𝜑-programs in one line")
+optSugar' :: Parser SugarType
+optSugar' = flag SALTY SWEET (long "sweet" <> help "Print result 𝜑-program using syntax sugar")
+
+optLineFormat :: [String] -> Parser LineFormat
+optLineFormat opts = flag MULTILINE SINGLELINE (long "flat" <> help (printf "Print result and intermediate (see %s option(s)) 𝜑-programs in one line" (intercalate ", " opts)))
+
+optLineFormat' :: Parser LineFormat
+optLineFormat' = flag MULTILINE SINGLELINE (long "flat" <> help "Print result 𝜑-program in one line")
+
+optOmitListing :: Parser Bool
+optOmitListing = switch (long "omit-listing" <> help "Omit full program listing in XMIR output")
+
+optOmitComments :: Parser Bool
+optOmitComments = switch (long "omit-comments" <> help "Omit comments in XMIR output")
 
 explainParser :: Parser Command
 explainParser =
@@ -186,49 +223,67 @@ explainParser =
 
 dataizeParser :: Parser Command
 dataizeParser =
-  CmdDataize
-    <$> ( OptsDataize
-            <$> optLogLevel
-            <*> optInputFormat
-            <*> optSugar
-            <*> optLineFormat
-            <*> optMaxDepth
-            <*> optMaxCycles
-            <*> optDepthSensitive
-            <*> optStepsDir
-            <*> argInputFile
-        )
+  let steps = ["--steps-dir"]
+   in CmdDataize
+        <$> ( OptsDataize
+                <$> optLogLevel
+                <*> optInputFormat
+                <*> optSugar steps
+                <*> optLineFormat steps
+                <*> optMaxDepth
+                <*> optMaxCycles
+                <*> optDepthSensitive
+                <*> optStepsDir
+                <*> argInputFile
+            )
 
 rewriteParser :: Parser Command
 rewriteParser =
-  CmdRewrite
-    <$> ( OptsRewrite
+  let opts = ["--sequence", "--steps-dir"]
+   in CmdRewrite
+        <$> ( OptsRewrite
+                <$> optLogLevel
+                <*> optRule
+                <*> optInputFormat
+                <*> optOutputFormat
+                <*> optSugar opts
+                <*> optLineFormat opts
+                <*> optNormalize
+                <*> optShuffle
+                <*> optOmitListing
+                <*> optOmitComments
+                <*> optDepthSensitive
+                <*> switch (long "nonumber" <> help "Turn off equation auto numbering in LaTeX rendering")
+                <*> option
+                  auto
+                  ( long "must"
+                      <> metavar "RANGE"
+                      <> help "Must-rewrite range (e.g., '3', '..5', '3..', '3..5'). Stops execution if number of rules applied is not in range. Use 0 to disable."
+                      <> value MtDisabled
+                      <> showDefaultWith show
+                  )
+                <*> optMaxDepth
+                <*> optMaxCycles
+                <*> switch (long "in-place" <> help "Edit file in-place instead of printing to output")
+                <*> switch (long "sequence" <> help "Result output contains all intermediate 𝜑-programs concatenated with EOL")
+                <*> optTarget
+                <*> optStepsDir
+                <*> argInputFile
+            )
+
+mergeParser :: Parser Command
+mergeParser =
+  CmdMerge
+    <$> ( OptsMerge
             <$> optLogLevel
-            <*> optRule
             <*> optInputFormat
-            <*> option (parseIOFormat "output") (long "output" <> metavar "FORMAT" <> help "Program output format (phi, xmir)" <> value PHI <> showDefault)
-            <*> optSugar
-            <*> optLineFormat
-            <*> optNormalize
-            <*> optShuffle
-            <*> switch (long "omit-listing" <> help "Omit full program listing in XMIR output")
-            <*> switch (long "omit-comments" <> help "Omit comments in XMIR output")
-            <*> optDepthSensitive
-            <*> option
-              auto
-              ( long "must"
-                  <> metavar "RANGE"
-                  <> help "Must-rewrite range (e.g., '3', '..5', '3..', '3..5'). Stops execution if number of rules applied is not in range. Use 0 to disable."
-                  <> value MtDisabled
-                  <> showDefaultWith show
-              )
-            <*> optMaxDepth
-            <*> optMaxCycles
-            <*> switch (long "in-place" <> help "Edit file in-place instead of printing to output")
-            <*> switch (long "sequence" <> help "Result output contains all intermediate 𝜑-programs concatenated with EOL")
+            <*> optOutputFormat
+            <*> optSugar'
+            <*> optLineFormat'
+            <*> optOmitListing
+            <*> optOmitComments
             <*> optTarget
-            <*> optStepsDir
-            <*> argInputFile
+            <*> many (argument str (metavar "[FILE]" <> help "Paths to input files"))
         )
 
 commandParser :: Parser Command
@@ -237,6 +292,7 @@ commandParser =
     ( command "rewrite" (info rewriteParser (progDesc "Rewrite the 𝜑-program"))
         <> command "dataize" (info dataizeParser (progDesc "Dataize the 𝜑-program"))
         <> command "explain" (info explainParser (progDesc "Explain rules in LaTeX format"))
+        <> command "merge" (info mergeParser (progDesc "Merge 𝜑-programs into single one by merging their top level formations"))
     )
 
 parserInfo :: ParserInfo Command
@@ -258,6 +314,7 @@ setLogLevel' cmd =
         CmdRewrite OptsRewrite {logLevel} -> logLevel
         CmdDataize OptsDataize {logLevel} -> logLevel
         CmdExplain OptsExplain {logLevel} -> logLevel
+        CmdMerge OptsMerge {logLevel} -> logLevel
    in setLogLevel level
 
 runCLI :: [String] -> IO ()
@@ -265,32 +322,35 @@ runCLI args = handle handler $ do
   cmd <- handleParseResult (execParserPure defaultPrefs parserInfo args)
   setLogLevel' cmd
   case cmd of
-    CmdRewrite opts@OptsRewrite {..} -> do
+    CmdRewrite OptsRewrite {..} -> do
       validateOpts
       logDebug (printf "Amount of rewriting cycles across all the rules: %d, per rule: %d" maxCycles maxDepth)
       input <- readInput inputFile
-      let xmirCtx = Just (XmirContext omitListing omitComments input)
       rules' <- getRules normalize shuffle rules
       program <- parseProgram input inputFormat
-      rewrittens <- rewrite' program rules' (context program xmirCtx)
+      let listing = if null rules' then const input else (\prog -> P.printProgram' prog (sugarType, UNICODE, flat))
+          xmirCtx = XmirContext omitListing omitComments listing
+          printProgCtx = PrintProgCtx sugarType flat xmirCtx nonumber
+      rewrittens <- rewrite' program rules' (context program printProgCtx)
       logDebug (printf "Printing rewritten 𝜑-program as %s" (show outputFormat))
-      progs <- printRewrittens xmirCtx rewrittens
+      progs <- printRewrittens printProgCtx rewrittens
       output targetFile progs
       where
         validateOpts :: IO ()
         validateOpts = do
-          when
-            (sugarType == SWEET && outputFormat == XMIR)
-            (throwIO (InvalidRewriteArguments "The --sweet and --output=xmir can't stay together"))
           when
             (inPlace && isNothing inputFile)
-            (throwIO (InvalidRewriteArguments "--in-place requires an input file"))
+            (throwIO (InvalidCLIArguments "--in-place requires an input file"))
           when
             (inPlace && isJust targetFile)
-            (throwIO (InvalidRewriteArguments "--in-place and --target cannot be used together"))
+            (throwIO (InvalidCLIArguments "--in-place and --target cannot be used together"))
+          when
+            (nonumber && outputFormat /= LATEX)
+            (throwIO (InvalidCLIArguments "The --nonumber option can stay together with --output=latex only"))
           validateMaxDepth maxDepth
           validateMaxCycles maxCycles
           validateMust must
+          validateXmirOptions outputFormat omitListing omitComments
         output :: Maybe FilePath -> String -> IO ()
         output target prog = case (inPlace, target, inputFile) of
           (True, _, Just file) -> do
@@ -304,7 +364,7 @@ runCLI args = handle handler $ do
           (False, Nothing, _) -> do
             logDebug "The option '--target' is not specified, printing to console..."
             putStrLn prog
-        context :: Program -> Maybe XmirContext -> RewriteContext
+        context :: Program -> PrintProgramContext -> RewriteContext
         context program ctx =
           RewriteContext
             program
@@ -314,12 +374,12 @@ runCLI args = handle handler $ do
             sequence
             buildTerm
             must
-            (saveStep stepsDir (ioToExtension outputFormat) (printProgram outputFormat (sugarType, flat) ctx))
-        printRewrittens :: Maybe XmirContext -> [Rewritten] -> IO String
+            (saveStep stepsDir (ioToExtension outputFormat) (printProgram outputFormat ctx))
+        printRewrittens :: PrintProgramContext -> [Rewritten] -> IO String
         printRewrittens ctx rewrittens
-          | outputFormat == LATEX = pure (rewrittensToLatex rewrittens (LatexContext sugarType flat))
-          | otherwise = mapM (printProgram outputFormat (sugarType, flat) ctx . program) rewrittens <&> intercalate "\n"
-    CmdDataize opts@OptsDataize {..} -> do
+          | outputFormat == LATEX = pure (rewrittensToLatex rewrittens (LatexContext sugarType flat nonumber))
+          | otherwise = mapM (printProgram outputFormat ctx . program) rewrittens <&> intercalate "\n"
+    CmdDataize OptsDataize {..} -> do
       validateOpts
       input <- readInput inputFile
       prog <- parseProgram input inputFormat
@@ -338,8 +398,8 @@ runCLI args = handle handler $ do
             maxCycles
             depthSensitive
             buildTerm
-            (saveStep stepsDir (ioToExtension PHI) (printProgram PHI (SWEET, MULTILINE) Nothing))
-    CmdExplain opts@OptsExplain {..} -> do
+            (saveStep stepsDir (ioToExtension PHI) (printProgram PHI (PrintProgCtx sugarType flat defaultXmirContext False)))
+    CmdExplain OptsExplain {..} -> do
       validateOpts
       rules' <- getRules normalize shuffle rules
       let latex = explainRules rules'
@@ -349,13 +409,38 @@ runCLI args = handle handler $ do
         validateOpts =
           when
             (null rules && not normalize)
-            (throwIO (InvalidRewriteArguments "Either --rule or --normalize must be specified"))
+            (throwIO (InvalidCLIArguments "Either --rule or --normalize must be specified"))
+    CmdMerge OptsMerge {..} -> do
+      validateOpts
+      inputs' <- traverse (readInput . Just) inputs
+      progs <- traverse (`parseProgram` inputFormat) inputs'
+      prog <- merge progs
+      let listing = const (P.printProgram' prog (sugarType, UNICODE, flat))
+          xmirCtx = XmirContext omitListing omitComments listing
+          printProgCtx = PrintProgCtx sugarType flat xmirCtx False
+      prog' <- printProgram outputFormat printProgCtx prog
+      output targetFile prog'
+      where
+        validateOpts :: IO ()
+        validateOpts = do
+          when
+            (null inputs)
+            (throwIO (InvalidCLIArguments "At least one input file must be specified for 'merge' command"))
+          validateXmirOptions outputFormat omitListing omitComments
   where
+    validateXmirOptions :: IOFormat -> Bool -> Bool -> IO ()
+    validateXmirOptions outputFormat omitListing omitComments = do
+      when
+        (outputFormat /= XMIR && omitListing)
+        (throwIO (InvalidCLIArguments "--omit-listing can be used only with --output-format=xmir"))
+      when
+        (outputFormat /= XMIR && omitComments)
+        (throwIO (InvalidCLIArguments "--omit-comments can be used only with --output-format=xmir"))
     validateIntArgument :: Integer -> (Integer -> Bool) -> String -> IO ()
     validateIntArgument num cmp msg =
       when
         (cmp num)
-        (throwIO (InvalidRewriteArguments msg))
+        (throwIO (InvalidCLIArguments msg))
     validateMaxDepth :: Integer -> IO ()
     validateMaxDepth depth = validateIntArgument depth (<= 0) "--max-depth must be positive"
     validateMaxCycles :: Integer -> IO ()
@@ -370,7 +455,7 @@ runCLI args = handle handler $ do
         (Just min, Just max)
           | min > max ->
               throwIO
-                ( InvalidRewriteArguments
+                ( InvalidCLIArguments
                     (printf "--must range invalid: minimum (%d) is greater than maximum (%d)" min max)
                 )
         _ -> pure ()
@@ -387,11 +472,10 @@ runCLI args = handle handler $ do
     parseProgram xmir XMIR = do
       doc <- parseXMIRThrows xmir
       xmirToPhi doc
-    printProgram :: IOFormat -> (SugarType, LineFormat) -> Maybe XmirContext -> Program -> IO String
-    printProgram PHI (sugar, line) _ prog = pure (P.printProgram' prog (sugar, UNICODE, line))
-    printProgram XMIR cfg Nothing prog = printProgram XMIR cfg (Just defaultXmirContext) prog
-    printProgram XMIR _ (Just ctx) prog = programToXMIR prog ctx <&> printXMIR
-    printProgram LATEX (sugar, line) _ prog = pure (programToLaTeX prog (LatexContext sugar line))
+    printProgram :: IOFormat -> PrintProgramContext -> Program -> IO String
+    printProgram PHI PrintProgCtx {..} prog = pure (P.printProgram' prog (sugar, UNICODE, line))
+    printProgram XMIR PrintProgCtx {..} prog = programToXMIR prog xmirCtx <&> printXMIR
+    printProgram LATEX PrintProgCtx {..} prog = pure (programToLaTeX prog (LatexContext sugar line nonumber))
     getRules :: Bool -> Bool -> [FilePath] -> IO [Y.Rule]
     getRules normalize shuffle rules = do
       ordered <-
