@@ -5,7 +5,7 @@
 
 module Functions (buildTerm) where
 
-import Ast
+import AST
 import Builder
 import Control.Exception (throwIO)
 import Control.Monad (replicateM, when)
@@ -20,36 +20,42 @@ import Matcher
 import Misc
 import Numeric (showHex)
 import Parser (parseAttributeThrows, parseNumberThrows)
-import Pretty
 import Random (randomString)
 import Regexp
 import System.Random (randomRIO)
 import Deps
 import Text.Printf (printf)
 import qualified Yaml as Y
+import Printer (printAttribute, printExpression, printExtraArg)
+import Logger (logDebug)
 
 buildTerm :: BuildTermFunc
-buildTerm "contextualize" = _contextualize
-buildTerm "scope" = _scope
-buildTerm "random-tau" = _randomTau
-buildTerm "dataize" = _dataize
-buildTerm "concat" = _concat
-buildTerm "sed" = _sed
-buildTerm "random-string" = _randomString
-buildTerm "size" = _size
-buildTerm "tau" = _tau
-buildTerm "string" = _string
-buildTerm "number" = _number
-buildTerm "sum" = _sum
-buildTerm "join" = _join
-buildTerm func = _unsupported func
+buildTerm func args subst prog = do
+  logDebug (printf "Building new term using '%s' function..." func)
+  buildTerm' func args subst prog
+
+buildTerm' :: BuildTermFunc
+buildTerm' "contextualize" = _contextualize
+buildTerm' "scope" = _scope
+buildTerm' "random-tau" = _randomTau
+buildTerm' "dataize" = _dataize
+buildTerm' "concat" = _concat
+buildTerm' "sed" = _sed
+buildTerm' "random-string" = _randomString
+buildTerm' "size" = _size
+buildTerm' "tau" = _tau
+buildTerm' "string" = _string
+buildTerm' "number" = _number
+buildTerm' "sum" = _sum
+buildTerm' "join" = _join
+buildTerm' func = _unsupported func
 
 argToBytes :: Y.ExtraArgument -> Subst -> Program -> IO Bytes
 argToBytes (Y.ArgBytes bytes) subst _ = buildBytesThrows bytes subst
 argToBytes (Y.ArgExpression expr) subst prog = do
   (TeBytes bts) <- _dataize [Y.ArgExpression expr] subst prog
   pure bts
-argToBytes arg _ _ = throwIO (userError (printf "Can't extract bytes from given argument: %s" (prettyExtraArg arg)))
+argToBytes arg _ _ = throwIO (userError (printf "Can't extract bytes from given argument: %s" (printExtraArg arg)))
 
 argToString :: Y.ExtraArgument -> Subst -> Program -> IO String
 argToString arg subst prog = argToBytes arg subst prog <&> btsToUnescapedStr
@@ -83,7 +89,7 @@ _randomTau args subst _ = do
       Y.ArgAttribute attr -> do
         attr' <- buildAttributeThrows attr subst
         rest' <- argsToAttrs rest
-        pure (prettyAttribute attr' : rest')
+        pure (printAttribute attr' : rest')
       Y.ArgBinding bd -> do
         bds <- buildBindingThrows bd subst
         rest' <- argsToAttrs rest
@@ -186,13 +192,13 @@ _string [Y.ArgExpression expr] subst _ = do
         ( userError
             ( printf
                 "Couldn't convert given expression to 'Φ̇.string' object, only 'Φ̇.number' or 'Φ̇.string' are allowed\n%s"
-                (prettyExpression' exp)
+                (printExpression exp)
             )
         )
   pure (TeExpression str)
 _string [Y.ArgAttribute attr] subst _ = do
   attr' <- buildAttributeThrows attr subst
-  pure (TeExpression (DataString (strToBts (prettyAttribute attr'))))
+  pure (TeExpression (DataString (strToBts (printAttribute attr'))))
 _string _ _ _ = throwIO (userError "Function string() requires exactly 1 argument as attribute or data expression (Φ̇.number or Φ̇.string)")
 
 _number :: BuildTermMethod
@@ -202,7 +208,7 @@ _number [Y.ArgExpression expr] subst _ = do
     DataString bts -> do
       num <- parseNumberThrows (btsToUnescapedStr bts)
       pure (TeExpression num)
-    _ -> throwIO (userError (printf "Function number() expects expression to be 'Φ̇.string', but got:\n%s" (prettyExpression' expr')))
+    _ -> throwIO (userError (printf "Function number() expects expression to be 'Φ̇.string', but got:\n%s" (printExpression expr')))
 _number _ _ _ = throwIO (userError "Function number() requires exactly 1 argument as 'Φ̇.string'")
 
 _sum :: BuildTermMethod
@@ -243,4 +249,4 @@ _join args subst _ = do
        in attr'
 
 _unsupported :: BuildTermFunc
-_unsupported func _ _ _ = throwIO (userError (printf "Function %s() is not supported or does not exist" func))
\ No newline at end of file
+_unsupported func _ _ _ = throwIO (userError (printf "Function %s() is not supported or does not exist" func))
