@@ -1,17 +1,19 @@
 {-# LANGUAGE DeriveAnyClass #-}
 {-# LANGUAGE DerivingStrategies #-}
+{-# LANGUAGE DuplicateRecordFields #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE RecordWildCards #-}
 
 -- SPDX-FileCopyrightText: Copyright (c) 2025 Objectionary.com
 -- SPDX-License-Identifier: MIT
 
-module Rewriter (rewrite, rewrite', RewriteContext (..)) where
+module Rewriter (rewrite, rewrite', RewriteContext (..), Rewritten (..)) where
 
-import Ast
+import AST
 import Builder
 import Control.Exception (Exception, throwIO)
 import Data.Foldable (foldlM)
+import Data.Functor ((<&>))
 import qualified Data.Map.Strict as M
 import Data.Maybe (catMaybes, fromMaybe, isJust)
 import qualified Data.Set as Set
@@ -21,7 +23,7 @@ import Matcher (MetaValue (MvAttribute, MvBindings, MvBytes, MvExpression), Subs
 import Misc (ensuredFile)
 import Must (Must (..), exceedsUpperBound, inRange)
 import Parser (parseProgram, parseProgramThrows)
-import Pretty (Encoding (UNICODE), PrintMode (SWEET), prettyAttribute, prettyBytes, prettyExpression, prettyExpression', prettyProgram, prettyProgram', prettySubsts)
+import Printer (printProgram)
 import Replacer (ReplaceProgramContext (ReplaceProgramContext), ReplaceProgramThrowsFunc, replaceProgramFastThrows, replaceProgramThrows)
 import Rule (RuleContext (RuleContext), matchProgramWithRule)
 import qualified Rule as R
@@ -29,11 +31,20 @@ import Text.Printf
 import Yaml (ExtraArgument (..))
 import qualified Yaml as Y
 
+data Rewritten = Rewritten
+  { program :: Program,
+    maybeRule :: Maybe Y.Rule
+  }
+
+instance Eq Rewritten where
+  left == right = program left == program right
+
 data RewriteContext = RewriteContext
   { _program :: Program,
     _maxDepth :: Integer,
     _maxCycles :: Integer,
     _depthSensitive :: Bool,
+    _sequence :: Bool,
     _buildTerm :: BuildTermFunc,
     _must :: Must,
     _saveStep :: SaveStepFunc
@@ -118,56 +129,72 @@ tryBuildAndReplaceFast (ExFormation pbds) (ExFormation rbds) substs ctx =
     hasMetaBindings = foldl (\acc bd -> acc || isMetaBinding bd) False
 tryBuildAndReplaceFast ptn res substs ctx = buildAndReplace' ptn res substs replaceProgramThrows ctx
 
-rewrite :: Program -> [Y.Rule] -> Set.Set Program -> Integer -> RewriteContext -> IO (Program, Set.Set Program)
-rewrite program [] progs _ _ = pure (program, progs)
-rewrite program (rule : rest) progs iteration ctx@RewriteContext {..} = do
-  (prog, _progs) <- _rewrite program 1 progs
-  rewrite prog rest _progs iteration ctx
+-- The function returns tuple (X, Y) where
+-- - X is sequence of programs; for more details about this sequence see description of rewrite' function
+-- - Y is Set of unique programs after each rule application. It allows to stop the rewriting if we're getting
+--   into loop and get back to program which we've already got before
+rewrite :: [Rewritten] -> [Y.Rule] -> Set.Set Program -> Integer -> RewriteContext -> IO ([Rewritten], Set.Set Program)
+rewrite rewrittens [] unique _ _ = pure (rewrittens, unique)
+rewrite rewrittens (rule : rest) unique iteration ctx@RewriteContext {..} = do
+  (rewrittens', unique') <- _rewrite rewrittens 1 unique
+  rewrite rewrittens' rest unique' iteration ctx
   where
-    _rewrite :: Program -> Integer -> Set.Set Program -> IO (Program, Set.Set Program)
-    _rewrite prog count progs' =
+    _rewrite :: [Rewritten] -> Integer -> Set.Set Program -> IO ([Rewritten], Set.Set Program)
+    _rewrite _rewrittens _count _unique =
       let ruleName = fromMaybe "unknown" (Y.name rule)
           ptn = Y.pattern rule
           res = Y.result rule
-       in if count - 1 == _maxDepth
+          Rewritten {..} = head _rewrittens
+       in if _count - 1 == _maxDepth
             then do
               logDebug (printf "Max amount of rewriting cycles (%d) for rule '%s' has been reached, rewriting is stopped" _maxDepth ruleName)
               if _depthSensitive
                 then throwIO (StoppedOnLimit "max-depth" _maxDepth)
-                else pure (prog, progs')
+                else pure (_rewrittens, _unique)
             else do
-              logDebug (printf "Starting rewriting cycle for rule '%s': %d out of %d" ruleName count _maxDepth)
-              matched <- R.matchProgramWithRule prog rule (RuleContext _program _buildTerm)
+              logDebug (printf "Starting rewriting cycle for rule '%s': %d out of %d" ruleName _count _maxDepth)
+              matched <- R.matchProgramWithRule program rule (RuleContext _program _buildTerm)
               if null matched
                 then do
                   logDebug (printf "Rule '%s' does not match, rewriting is stopped" ruleName)
-                  pure (prog, progs')
+                  pure (_rewrittens, _unique)
                 else do
                   logDebug (printf "Rule '%s' has been matched, applying..." ruleName)
-                  prog' <- tryBuildAndReplaceFast ptn res matched (ReplaceProgramContext prog _maxDepth)
-                  if prog == prog'
+                  prog <- tryBuildAndReplaceFast ptn res matched (ReplaceProgramContext program _maxDepth)
+                  if program == prog
                     then do
                       logDebug (printf "Applied '%s', no changes made" ruleName)
-                      pure (prog, progs')
+                      pure (_rewrittens, _unique)
                     else
-                      if Set.member prog' progs
-                        then throwIO (LoopingRewriting (prettyProgram' prog' SWEET UNICODE) ruleName count)
+                      if Set.member prog _unique
+                        then throwIO (LoopingRewriting (printProgram prog) ruleName _count)
                         else do
                           logDebug
                             ( printf
                                 "Applied '%s' (%d nodes -> %d nodes)"
                                 ruleName
+                                (countNodes program)
                                 (countNodes prog)
-                                (countNodes prog')
                             )
-                          _saveStep prog' (((iteration - 1) * _maxDepth) + count)
-                          _rewrite prog' (count + 1) (Set.insert prog' progs)
+                          _saveStep prog (((iteration - 1) * _maxDepth) + _count)
+                          _rewrite (rewriteSequence prog) (_count + 1) (Set.insert prog _unique)
+      where
+        rewriteSequence :: Program -> [Rewritten]
+        rewriteSequence _prog
+          | _sequence =
+              let Rewritten {..} : rest = _rewrittens
+               in Rewritten _prog Nothing : Rewritten program (Just rule) : rest
+          | otherwise = [Rewritten _prog Nothing]
 
-rewrite' :: Program -> [Y.Rule] -> RewriteContext -> IO Program
-rewrite' prog rules ctx = _rewrite prog 1 Set.empty
+-- The function accepts single program but returns sequence of programs
+-- If RewriteContext has _sequence == True - all the intermediate
+-- programs after each rule application are included into sequence
+-- Otherwise sequence contains only one program
+rewrite' :: Program -> [Y.Rule] -> RewriteContext -> IO [Rewritten]
+rewrite' prog rules ctx = _rewrite [Rewritten prog Nothing] 1 Set.empty <&> reverse
   where
-    _rewrite :: Program -> Integer -> Set.Set Program -> IO Program
-    _rewrite prog count progs = do
+    _rewrite :: [Rewritten] -> Integer -> Set.Set Program -> IO [Rewritten]
+    _rewrite rewrittens count unique = do
       let cycles = _maxCycles ctx
           must = _must ctx
           current = count - 1
@@ -179,14 +206,14 @@ rewrite' prog rules ctx = _rewrite prog 1 Set.empty
               logDebug (printf "Max amount of rewriting cycles for all rules (%d) has been reached, rewriting is stopped" cycles)
               if _depthSensitive ctx
                 then throwIO (StoppedOnLimit "max-cycles" cycles)
-                else pure prog
+                else pure rewrittens
             else do
               logDebug (printf "Starting rewriting cycle for all rules: %d out of %d" count cycles)
-              (rewritten, progs') <- rewrite prog rules progs count ctx
-              if rewritten == prog
+              (rewrittens', unique') <- rewrite rewrittens rules unique count ctx
+              if head rewrittens' == head rewrittens
                 then do
                   logDebug "Rewriting is stopped since it has no effect"
                   if not (inRange must current)
                     then throwIO (MustBeGoing must current)
-                    else pure rewritten
-                else _rewrite rewritten (count + 1) progs'
+                    else pure rewrittens'
+                else _rewrite rewrittens' (count + 1) unique'
